import React from 'react';
import { IRecommendedVersion, Room } from "matrix-js-sdk/src/models/room";
import { MatrixEvent } from "matrix-js-sdk/src/models/event";
import { ISearchResults } from 'matrix-js-sdk/src/@types/search';
import { EventTimeline } from 'matrix-js-sdk/src/models/event-timeline';
import { CallState } from "matrix-js-sdk/src/webrtc/call";
import { MatrixError } from 'matrix-js-sdk/src/http-api';
import ResizeNotifier from '../../utils/ResizeNotifier';
import { Layout } from "../../settings/enums/Layout";
import { TimelineRenderingType } from "../../contexts/RoomContext";
import MatrixClientContext, { MatrixClientProps } from "../../contexts/MatrixClientContext";
import { E2EStatus } from '../../utils/ShieldUtils';
import { IMatrixClientCreds } from "../../MatrixClientPeg";
import { SearchScope } from "../views/rooms/SearchBar";
import { XOR } from "../../@types/common";
import { IOOBData, IThreepidInvite } from "../../stores/ThreepidInviteStore";
import { IOpts } from "../../createRoom";
import EditorStateTransfer from "../../utils/EditorStateTransfer";
interface IRoomProps extends MatrixClientProps {
    threepidInvite: IThreepidInvite;
    oobData?: IOOBData;
    resizeNotifier: ResizeNotifier;
    justCreatedOpts?: IOpts;
    forceTimeline?: boolean;
    onRegistered?(credentials: IMatrixClientCreds): void;
}
declare enum MainSplitContentType {
    Timeline = 0,
    MaximisedWidget = 1,
    Video = 2
}
export interface IRoomState {
    room?: Room;
    roomId?: string;
    roomAlias?: string;
    roomLoading: boolean;
    peekLoading: boolean;
    shouldPeek: boolean;
    membersLoaded: boolean;
    initialEventId?: string;
    initialEventPixelOffset?: number;
    isInitialEventHighlighted?: boolean;
    initialEventScrollIntoView?: boolean;
    replyToEvent?: MatrixEvent;
    numUnreadMessages: number;
    searchTerm?: string;
    searchScope?: SearchScope;
    searchResults?: XOR<{}, ISearchResults>;
    searchHighlights?: string[];
    searchInProgress?: boolean;
    callState?: CallState;
    canPeek: boolean;
    canSelfRedact: boolean;
    showApps: boolean;
    isPeeking: boolean;
    showRightPanel: boolean;
    roomLoadError?: MatrixError;
    joining: boolean;
    atEndOfLiveTimeline?: boolean;
    showTopUnreadMessagesBar: boolean;
    statusBarVisible: boolean;
    upgradeRecommendation?: IRecommendedVersion;
    canReact: boolean;
    canSendMessages: boolean;
    tombstone?: MatrixEvent;
    resizing: boolean;
    layout: Layout;
    lowBandwidth: boolean;
    alwaysShowTimestamps: boolean;
    showTwelveHourTimestamps: boolean;
    readMarkerInViewThresholdMs: number;
    readMarkerOutOfViewThresholdMs: number;
    showHiddenEvents: boolean;
    showReadReceipts: boolean;
    showRedactions: boolean;
    showJoinLeaves: boolean;
    showAvatarChanges: boolean;
    showDisplaynameChanges: boolean;
    matrixClientIsReady: boolean;
    showUrlPreview?: boolean;
    e2eStatus?: E2EStatus;
    rejecting?: boolean;
    rejectError?: Error;
    hasPinnedWidgets?: boolean;
    mainSplitContentType?: MainSplitContentType;
    wasContextSwitch?: boolean;
    editState?: EditorStateTransfer;
    timelineRenderingType: TimelineRenderingType;
    threadId?: string;
    liveTimeline?: EventTimeline;
    narrow: boolean;
}
export declare class RoomView extends React.Component<IRoomProps, IRoomState> {
    private readonly dispatcherRef;
    private readonly roomStoreToken;
    private settingWatchers;
    private unmounted;
    private permalinkCreators;
    private searchId;
    private roomView;
    private searchResultsPanel;
    private messagePanel;
    private roomViewBody;
    static contextType: React.Context<import("matrix-js-sdk/src/client").MatrixClient>;
    context: React.ContextType<typeof MatrixClientContext>;
    constructor(props: IRoomProps, context: React.ContextType<typeof MatrixClientContext>);
    private onIsResizing;
    private onWidgetStoreUpdate;
    private onWidgetEchoStoreUpdate;
    private onWidgetLayoutChange;
    private checkWidgets;
    private getMainSplitContentType;
    private onRoomViewStoreUpdate;
    private getRoomId;
    private getPermalinkCreatorForRoom;
    private stopAllPermalinkCreators;
    private setupRoom;
    private shouldShowApps;
    componentDidMount(): void;
    shouldComponentUpdate(nextProps: any, nextState: any): boolean;
    componentDidUpdate(): void;
    componentWillUnmount(): void;
    private onRightPanelStoreUpdate;
    private onPageUnload;
    private onReactKeyDown;
    private onCallState;
    private onAction;
    private onLocalRoomEvent;
    private onRoomTimeline;
    private onEventDecrypted;
    private handleEffects;
    private onRoomName;
    private onKeyBackupStatus;
    canResetTimeline: () => boolean;
    private onRoomLoaded;
    private onRoomTimelineReset;
    private getRoomTombstone;
    private calculateRecommendedVersion;
    private loadMembersIfJoined;
    private calculatePeekRules;
    private updatePreviewUrlVisibility;
    private onRoom;
    private onDeviceVerificationChanged;
    private onUserVerificationChanged;
    private onCrossSigningKeysChanged;
    private updateE2EStatus;
    private onUrlPreviewsEnabledChange;
    private onRoomStateEvents;
    private onRoomStateUpdate;
    private onMyMembership;
    private updatePermissions;
    private updateRoomMembers;
    private checkDesktopNotifications;
    private updateDMState;
    private onSearchResultsFillRequest;
    private onInviteClick;
    private onJoinButtonClicked;
    private onMessageListScroll;
    private resetJumpToEvent;
    private injectSticker;
    private onSearch;
    private handleSearchResult;
    private getSearchResultTiles;
    private onCallPlaced;
    private onAppsClick;
    private onForgetClick;
    private onRejectButtonClicked;
    private onRejectAndIgnoreClick;
    private onRejectThreepidInviteButtonClicked;
    private onSearchClick;
    private onCancelSearchClick;
    private jumpToLiveTimeline;
    private jumpToReadMarker;
    private forgetReadMarker;
    private updateTopUnreadMessagesBar;
    private getScrollState;
    private onStatusBarVisible;
    private onStatusBarHidden;
    /**
     * called by the parent component when PageUp/Down/etc is pressed.
     *
     * We pass it down to the scroll panel.
     */
    handleScrollKey: (ev: any) => void;
    /**
     * get any current call for this room
     */
    private getCallForRoom;
    private gatherTimelinePanelRef;
    private getOldRoom;
    getHiddenHighlightCount(): number;
    onHiddenHighlightsClick: () => void;
    private get messagePanelClassNames();
    private onFileDrop;
    private onMeasurement;
    private get viewsLocalRoom();
    private get permalinkCreator();
    private renderLocalRoomCreateLoader;
    private renderLocalRoomView;
    render(): JSX.Element;
}
declare const RoomViewWithMatrixClient: React.ForwardRefExoticComponent<Omit<IRoomProps, "mxClient"> & React.RefAttributes<React.Component<IRoomProps, any, any>>>;
export default RoomViewWithMatrixClient;
